# coding: utf-8
from piaf.agent import Agent
from piaf.comm import AID
from piaf.comm import ACLMessage, Performative
from piaf.behavior import Behavior, FSMBehavior


class SendMsgBehavior(Behavior):
    """A simple behavior that sends messages."""

    def __init__(self, agent, msg, *args, **kwargs):
        """
        Create a new :class:`SendMsgBehavior`.

        :param agent: the agent running this behavior
        :param msg: the message to send
        """
        super().__init__(agent, *args, **kwargs)
        self.msg = msg

    async def action(self):
        """Send the message and increase the agent send counter."""
        self.agent.send(self.msg)
        self.agent.send_count += 1

    def result(self):
        """Get the counter's value."""
        return self.agent.send_count


class RcvMsgBehavior(Behavior):
    """A simple behavior that receives messages."""

    async def action(self):
        """Receive a message, display it and increase the agent receive counter."""
        msg = await self.agent.receive()
        self.agent.rcv_count += 1
        self.agent.logger.info(
            "[%s] Received %s", self.agent.aid.short_name, msg.acl_message.content
        )

    def result(self):
        """Return the agent receive counter."""
        return self.agent.rcv_count


class TerminateAgentBehavior(Behavior):
    """A simple behavior that terminates the agent."""

    async def action(self):
        """Terminate the agent."""
        await self.agent.quit()


class PingAgent(Agent):
    """
    The Ping agent is an agent that sends "PING" messages.

    This agent uses a FSM::

        SEND ----> RECEIVE --- (rcv > 10) --> END
          ^                 |
          ---( rcv <= 10) ---

    """

    def __init__(self, aid, platform):
        super().__init__(aid, platform)
        self.send_count = 0
        self.rcv_count = 0

        pong = AID("Pong@localhost", ("memory://localhost/Pong",))
        msg = (
            ACLMessage.Builder()
            .receiver(pong)
            .content("PING")
            .performative(Performative.INFORM)
            .build()
        )

        bhv = FSMBehavior(self)
        bhv.add_state("SEND", SendMsgBehavior, args=(msg,), final=False)
        bhv.add_state("RCV", RcvMsgBehavior)
        bhv.add_state("END", TerminateAgentBehavior, final=True)

        # Careful: order matter
        bhv.add_transition("RCV", "END", lambda r: r >= 10)
        bhv.add_transition("SEND", "RCV", lambda r: True)
        bhv.add_transition("RCV", "SEND", lambda r: True)

        bhv.set_initial_state("SEND")
        self.add_behavior(bhv)


class PongAgent(Agent):
    """
    Same as Ping agent, but slightly different::

        RECEIVE ----> SEND --- (send > 10) --> END
          ^                 |
          ---( send <= 10) ---

    """

    def __init__(self, aid, platform):
        super().__init__(aid, platform)
        self.send_count = 0
        self.rcv_count = 0

        pong = AID("Ping@localhost", ("memory://localhost/Ping",))
        msg = (
            ACLMessage.Builder()
            .receiver(pong)
            .content("PONG")
            .performative(Performative.INFORM)
            .build()
        )

        bhv = FSMBehavior(self)
        bhv.add_state("SEND", SendMsgBehavior, args=(msg,), final=False)
        bhv.add_state("RCV", RcvMsgBehavior)
        bhv.add_state("END", TerminateAgentBehavior, final=True)

        # Careful: order matter
        bhv.add_transition("SEND", "END", lambda r: r >= 10)
        bhv.add_transition("SEND", "RCV", lambda r: True)
        bhv.add_transition("RCV", "SEND", lambda r: True)

        bhv.set_initial_state("RCV")
        self.add_behavior(bhv)


if __name__ == "__main__":
    import logging
    from piaf.launcher import AgentDescription, PlatformLauncher

    # Configure logging level and handler to see things
    logging.getLogger().setLevel(logging.DEBUG)
    logging.getLogger().addHandler(logging.StreamHandler())

    # Create platform launcher
    launcher = PlatformLauncher("localhost")

    # Now we can add our agents
    launcher.add_agent(AgentDescription("Ping", PingAgent))
    launcher.add_agent(AgentDescription("Pong", PongAgent))

    # The program will run until you hit Ctrl+C
    launcher.run()



"""
Part of the "two-platforms" test.

This module contains everything to setup and run the first platform. Once invoked, it will
create a platform named "ptf1" and one agent, called "my-agent".

The agent will try to send a message to an agent on the second platform. To do so, it will
first contact the ptf2 AMS periodically until the AMS tells the agent is up and running. Then
it can send the message!

.. note: In order to run this example, you will have to setup an AMQP server. Using Docker and RabbitMQ,
    you can easily launch the example: `docker run -it --rm --name rabbitmq -p 5672:5672 -p 15672:15672 rabbitmq:3.9-management-alpine`

Once the server is ready, just run `python ptf2.py` in one terminal and then `python ptf1.py` in another. If the example
works, you should see in the ptf2 terminal the message sent by the ptf1 agent.
"""
from __future__ import annotations
import asyncio
from typing import Any

from piaf.agent import Agent, AgentState
from piaf.ptf import AgentPlatformFacade
from piaf.service import AMSService, AMSAgentDescription
from piaf.behavior import Behavior, FSMBehavior, SuicideBehavior
from piaf.comm import (
    AID,
    MT_AND,
    MT_CONVERSATION_ID,
    MT_OR,
    MT_PERFORMATIVE,
    ACLMessage,
    Message,
    Performative,
)


class SendMsgBehavior(Behavior):
    """A simple behavior that sends messages."""

    def __init__(self, agent: Agent, msg: Message) -> None:
        """
        Create a new :class:`SendMsgBehavior`.

        :param agent: the agent running this behavior
        :param msg: the message to send
        """
        super().__init__(agent)
        self.msg = msg

    async def action(self):
        """Send the message."""
        self.agent.send(self.msg)


class FindOtherAgent(Behavior):
    """
    A simple behavior which runs until the other agent (the one living on the other
    platform) is available.
    """

    def __init__(self, agent: Agent, ams: AID) -> None:
        """
        Initialize the behavior.

        It expects the agent (like any behavior) and the AID of the other agent.
        :param agent: the agent
        :param aid: other agent's AID
        """
        super().__init__(agent)
        self.ams: AID = ams
        self.found = True

    async def action(self) -> None:
        # Contact other platform to ask if the other agent is ready
        self.agent.logger.info("Sending lookup request.")
        conv = "find_agent"
        req = (
            ACLMessage.Builder()
            .performative(Performative.REQUEST)
            .receiver(self.ams)
            .conversation_id(conv)
            .content(
                [
                    AMSService.SEARCH_FUNC,
                    AMSAgentDescription(
                        name=AID(f"other@{self.ams.hap_name}"), state=AgentState.ACTIVE
                    ),
                ]
            )
            .build()
        )
        self.agent.send(req)

        # Read response
        reply = await self.agent.receive(
            MT_AND(
                MT_CONVERSATION_ID(conv),
                MT_OR(
                    MT_PERFORMATIVE(Performative.AGREE),
                    MT_PERFORMATIVE(Performative.REFUSE),
                    MT_PERFORMATIVE(Performative.FAILURE),
                ),
            )
        )
        if reply.acl_message.performative != Performative.AGREE:
            self.logger.error("AMS refused my request :(")
            self.found = False
            await asyncio.sleep(1)
            return

        self.agent.logger.info("AMS replied with AGREE")
        resp = await self.agent.receive(MT_CONVERSATION_ID(conv))

        # If not found, then sleep a bit and set found to false
        if len(resp.acl_message.content[1]) == 0:
            self.agent.logger.info("Other agent is not ready, sleep a bit ...")
            await asyncio.sleep(1)
            self.found = False
            return

    def result(self) -> Any:
        """
        Return if the AMS replied with the other agent's AID.

        :return: `True` if the other agent is up and ready
        """
        return self.found


class MyAgent(Agent):
    """
    The agent that will query the remote platform.

    Internally, it uses an FSMBehavior to execute the :class:`FindOtherAgent`,
    :class:`SendMsgBehavior` and `TerminateAgentBehavior` in the right order.
    """

    def __init__(self, aid: AID, platform: AgentPlatformFacade, other: AID):
        """
        Initialize a new instance of the agent.

        :param aid: agent's aid
        :param platform: the platform on which the agent will run
        :param other: the other agent's AID
        """
        super().__init__(aid, platform)

        msg = (
            ACLMessage.Builder()
            .performative(Performative.INFORM)
            .receiver(other)
            .content("Hello from PTF1!")
            .build()
        )

        fsm = FSMBehavior(self)
        self.add_behavior(fsm)

        fsm.add_state(
            "FIND", FindOtherAgent, args=[AID(f"ams@ptf2", ["amqp://ptf2/acc"])]
        )
        fsm.add_state("SEND", SendMsgBehavior, args=[msg])
        fsm.add_state("DEATH", SuicideBehavior, final=True)
        fsm.set_initial_state("FIND")

        fsm.add_transition("FIND", "FIND", lambda found: not found)
        fsm.add_transition("FIND", "SEND", lambda found: found)
        fsm.add_transition("SEND", "DEATH", lambda e: True)


if __name__ == "__main__":
    import logging
    from piaf.launcher import AgentDescription, PlatformLauncher
    from piaf.comm.mtp.amqp import AMQPMessageTransportProtocol

    # Configure logging level and handler to see things
    logging.getLogger().setLevel(logging.INFO)
    logging.getLogger().addHandler(logging.StreamHandler())

    # Create the platform launcher
    launcher = PlatformLauncher("ptf1")

    # Register MTP. Since we want to connect this platform to another one, we need an MTP.
    # Here we are going to use the AMQPMessageTransferProtocol, which relies on AMQP 0.9.1 protocol.
    launcher.add_mtp(AMQPMessageTransportProtocol("amqp://guest:guest@localhost/"))

    # Now we can add our agent. The other agent AID is hardcode but we could use
    # the DF agent in a real case scenario to retrieve it.
    launcher.add_agent(
        AgentDescription(
            "my-agent",
            MyAgent,
            kwargs={"other": AID("other@ptf2", ["amqp://ptf2/acc"])},
        )
    )

    # The program will run until you hit Ctrl+C
    launcher.run()



# coding: utf-8
"""
This module contains every related to the Message Transport System.

The MTS is separated in different components:

* :class:`MessageTransferHandler`, which are components designed to handle
  communication using a specific protocol
* :class:`PlayloadParser`, which are responsible to dump and load
  :class:`piaf.comm.ACLMessage` to/from concrete representations
* the :class:`AgentCommunicationChannel` component, which provides message forwarding capability

class:`MessageTransferHandler` and class:`PlayloadParser` are interfaces that allow
you to create your own components (and thus use your own transport protocol or playload
representation).
"""
from __future__ import annotations

import abc
import asyncio
import copy
import itertools
import logging
import urllib.parse
from asyncio.futures import Future
from datetime import datetime
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Dict,
    Iterable,
    List,
    NamedTuple,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
)

import piaf.comm
import piaf.exceptions as ex
from piaf.comm import AID, ACLMessage, Envelope, Message, Performative, ReceivedObject
from piaf.audit import ACC_TOPIC, AGENTS_PARENT_TOPIC, Event, Topic

if TYPE_CHECKING:
    from piaf.ptf import AgentPlatform

from piaf.exceptions import DuplicatedNameException

__all__ = [
    "MessageTransportProtocol",
    "PlayloadParser",
    "AgentCommunicationChannel",
    "MailBox",
    "MessageContext",
]


class MailBox:
    """Agent's mailbox."""

    def __init__(self):
        """Create a new mailbox."""
        self._queue: List[Message] = []
        self._sync = asyncio.Condition()

    async def get(
        self, template: piaf.comm.MessageTemplate = piaf.comm.MT_ALL()
    ) -> piaf.comm.Message:
        """
        Retrieve the first matching message in the mail box.

        This method blocks until such message is inserted with the method :meth:`put`.
        Note that the message is removed from the queue so two successive calls to
        :meth:`get` (or :meth:`get_nowait`) will end up with two different messages.
        Template is applied on the :class:`ACLMessage`.
        """
        async with self._sync:
            msg = None
            while msg is None:
                msg = self._lookup_for_matching_message(template)
                if msg is None:
                    await self._sync.wait()
            return msg

    def _lookup_for_matching_message(
        self, template: piaf.comm.MessageTemplate = piaf.comm.MT_ALL()
    ) -> Union[None, piaf.comm.Message]:
        """
        Lookup the queue for a message matching the provided template.

        If no such message exists, this method will return ``None``. This method
        doesn't take the lock so callers must lock the call themselves.
        Template is applied on the :class:`ACLMessage`.
        """
        for i in range(len(self._queue)):
            if template.apply(self._queue[i].acl_message):
                return self._queue.pop(i)
        return None

    async def get_nowait(
        self, template: piaf.comm.MessageTemplate = piaf.comm.MT_ALL()
    ) -> Union[None, piaf.comm.Message]:
        """
        Retrieve the first matching message in the mail box.

        Unlike :meth:`get`, this method returns ``None`` if no matching message is
        found. Otherwise, the behavior is similar to :meth:`get`.
        """
        async with self._sync:
            return self._lookup_for_matching_message(template)

    async def put(self, item: piaf.comm.Message) -> None:
        """Put the provided message in this mailbox."""
        async with self._sync:
            self._queue.append(item)
            self._sync.notify_all()


class MailBoxDelegate:
    """
    Delegation to internal agent or service mailbox.

    Restrict access to agent/service mailbox. Only exposed method is :meth:`MailBox.put`.
    """

    def __init__(self, delegate: MailBox) -> None:
        """
        Create a new :class:`MailBoxDelegate` object.

        :param delegate: agent's mailbox
        """
        self._delegate: MailBox = delegate

    async def put(self, item: piaf.comm.Message) -> None:
        """Put the provided message in this mailbox."""
        await self._delegate.put(item)


class PlayloadParser(metaclass=abc.ABCMeta):
    """Abstract playload parser. Use concrete classes to load or dump playload."""

    _PARSERS: ClassVar[Dict[str, Type[PlayloadParser]]] = {}
    _COMPONENT_NAME: ClassVar[str] = ""

    def __init_subclass__(cls, cmp_name: str):
        super().__init_subclass__()
        PlayloadParser._PARSERS[cmp_name] = cls
        cls._COMPONENT_NAME = cmp_name

    @classmethod
    def get_parser(cls, cmp_name: str):
        return cls._PARSERS[cmp_name]

    @abc.abstractmethod
    def dump(self, data: Any, encoding: str = "utf-8") -> bytearray:
        """Dump the provided data into a :class:`bytearray`.

        :param data: the playload to dump
        :param encoding: how string are encoded
        :return: the playload dumped into the bytearray
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def load(self, data: bytearray, encoding: str = "utf-8") -> Any:
        """Load a playload from the provided bytearray.

        :param data: the bytearray
        :param encoding: how strings are encoded
        :return: a playload
        """
        raise NotImplementedError()


class AgentCommunicationChannel:
    """
    :class:`AgentCommunicationChannel` is an entity providing the messaging service to the AP agents.

    See http://fipa.org/specs/fipa00067/ for more information.
    """

    _ID = itertools.count(0)
    _id_lock = asyncio.Lock()

    @classmethod
    async def _next_id(cls):
        async with cls._id_lock:
            return next(cls._ID)

    def __init__(self, ptf: AgentPlatform) -> None:
        """Create a new ACC without any registered agent or service."""
        self._knows: Dict[str, MailBox] = {}
        self._tasks: List[Future] = []
        self._ptf: AgentPlatform = ptf
        self._mtps: Dict[str, MessageTransportProtocol] = {}
        self.logger = logging.getLogger(type(self).__name__)

    @property
    def schemes(self) -> Tuple[str, ...]:
        """Get the list of supported schemes."""
        return tuple(self._mtps)

    @property
    def ptf_name(self) -> str:
        """Get the platform name."""
        return self._ptf.name

    def register_agent_or_service(self, name: str, mailbox: MailBox) -> None:
        """
        Register an agent or a service to this ACC.

        If registration is successful, any message received by the ACC targeting the registered agent or service will
        be delivered to that agent or service using the provided mailbox.

        :raise DuplicatedNameException: an agent or a service with the provided name already exists.
        """
        if name in self._knows:
            raise DuplicatedNameException(name)

        self._knows[name] = mailbox

    def forget_agent_or_service(self, name: str) -> None:
        """
        Forget an agent or a service.

        The provided name will be forgotten and messages will no longer be delivered. No-op if the name is unknown.
        """
        if name in self._knows:
            del self._knows[name]

    async def register_mtp(self, mtp: MessageTransportProtocol) -> None:
        """
        Register a new MTP and start it.

        Once registered, the ACC will be able to use the MTP to send messages. ACC doesn't allow duplicated MTPs (ie,
        MTP providing the same scheme). If you try to register a MTP with an already registered scheme, this method
        will raise an exception.

        :raise DuplicatedSchemeException: if you try to register an MTP handling an already-register scheme.
        """
        if mtp.scheme in self._mtps:
            raise ex.DuplicatedSchemeException(mtp.scheme)

        self._mtps[mtp.scheme] = mtp
        mtp.acc = self
        await mtp.start()

        await self._ptf.evt_manager.publish(
            Event("acc", "mtp_registration", mtp.scheme), ACC_TOPIC
        )

    @property
    def addresses(self) -> Tuple[str, ...]:
        """Get all addresses this ACC can be contacted with."""
        return tuple(mtp.address for mtp in self._mtps.values())

    def forward(self, msg: Message, context: MessageContext = None):
        """
        Forward the provided message.

        Rules here.
        """
        task = asyncio.create_task(self._handle_message(msg, context))
        self._tasks.append(task)

    async def stop(self) -> None:
        """
        Stop the ACC.

        This will cancel all running tasks and stop all running MTPs.
        """
        for task in self._tasks:
            task.cancel()

        if self._tasks:
            for result in await asyncio.gather(*self._tasks, return_exceptions=True):
                if (
                    result is not None
                ):  # Exception occurred (result is None if success, see _handle_message)
                    self.logger.exception(
                        "[ACC] Message handling exception.", exc_info=result
                    )

        for mtp in self._mtps.values():
            await mtp.stop()

    async def _handle_message(
        self, msg: Message, context: MessageContext = None
    ) -> None:
        """
        Asynchronous task to take care of a message.

        :param msg: the message to handle
        :param context: optional context linked to the message
        """
        envelope = msg.envelope
        f_update: Dict[str, Any] = {
            Envelope.Builder.RECEIVED: await self._fill_received_field(context),
        }

        # Generate intended-receiver from the to field then re-run method
        if envelope.intended_receiver is None:
            f_update[Envelope.Builder.INTENDED_RECEIVER] = (
                envelope.to if isinstance(envelope.to, Iterable) else (envelope.to,)
            )
            envelope.add_revision(f_update)
            self.forward(msg, context)

        # Read intended-receiver field and dispatch
        else:
            local, others = self._retrieve_local_receivers(envelope.intended_receiver)

            if others:
                for aid in others:
                    f_update[Envelope.Builder.INTENDED_RECEIVER] = (aid,)
                    cpy: Message = copy.deepcopy(msg)
                    cpy.envelope.add_revision(f_update)
                    await self._send_via_mts(aid, cpy)

            if local:
                f_update[Envelope.Builder.INTENDED_RECEIVER] = local
                msg.envelope.add_revision(f_update)
                for aid in local:
                    cpy: Message = copy.deepcopy(msg)
                    cpy.envelope.add_revision(f_update)
                    await self._send_locally(aid, cpy)

    async def _fill_received_field(
        self, context: MessageContext = None
    ) -> ReceivedObject:
        """
        Generate new content for the 'received' field  with this ACC information.

        :param context: message context. Optional.
        """
        by = (
            f"memory://{self._ptf.name}/acc" if context is None else context.via.address
        )
        rcv_obj = ReceivedObject(
            by=by,
            date=datetime.now(),
            id=await AgentCommunicationChannel._next_id(),
        )

        if context is not None:
            rcv_obj.from_ = context.from_
            rcv_obj.via = context.via.type

        return rcv_obj

    async def _send_via_mts(self, aid: AID, msg: Message):
        """
        Send a message through available :class:`MessageTransportProtocol`.

        :param aid: AID of the receiver
        :param msg: Message to send
        """
        success = False
        for address in aid.addresses:
            scheme = self._retrieve_scheme_from_url(address)
            if scheme in self._mtps:
                success = await self._mtps[scheme].send(msg, address)

            if success:
                break

        # If still not successful, try name resolution
        # Not implemented yet
        if not success:
            pass

        # Nothing worked, send failure message
        if not success:
            await self._send_failure_message(msg)

    async def _send_locally(self, aid: AID, msg: Message) -> None:
        """
        Send a message to a local agent.

        :param aid: agent's AID
        :param msg: message to send
        """
        if aid.short_name in self._knows:
            await self._knows[aid.short_name].put(msg)

            # Fire msg reception event
            acl_msg: ACLMessage = copy.deepcopy(msg.acl_message)
            await self._ptf.evt_manager.publish(
                Event(aid.name, "message_reception", acl_msg),
                Topic.resolve_from_parent_topic(
                    AGENTS_PARENT_TOPIC, f"{aid.short_name}.messages"
                ),
            )
        else:
            await self._send_failure_message(msg)

    def _retrieve_scheme_from_url(self, url: str):
        """Retrieve scheme from provided URL."""
        return urllib.parse.urlparse(url).scheme

    def _retrieve_local_receivers(
        self, aids: Sequence[AID]
    ) -> Tuple[Iterable[AID], Iterable[AID]]:
        """
        Retrieve local receivers.

        Returns a tuple containing two sequences: the first one is an iterable of local AIDs, the second one contains
        external AIDs.

        :param aids: Iterable of AIDs containing potentially local AIDs
        """
        local = []
        others = []
        for aid in aids:
            if aid.hap_name == self._ptf.name:
                local.append(aid)
            else:
                others.append(aid)
        return local, others

    async def _send_failure_message(self, msg: Message) -> None:
        """
        Build and send a failure message from the provided message.

        Will extract:

        - conversation_id
        - reply_with
        - sender

        If those information are not provided, abort message sending and log error.
        """

        try:
            reply_msg = (
                ACLMessage.Builder()
                .performative(Performative.FAILURE)
                .receiver(msg.acl_message.sender)
                .conversation_id(msg.acl_message.conversation_id)
                .content([msg.acl_message.content, f"Unreachable agent: {msg.acl_message.receiver}"])
                .build()
            )
            reply_msg.in_reply_to = msg.acl_message.reply_with

            reply_envlp = (
                Envelope.Builder()
                .from_(AID(f"ams@{self._ptf.name}", addresses=self.addresses))
                .to((msg.acl_message.sender,))
                .acl_representation("fipa.acl.rep.string.std")
                .date(datetime.now())
                .build()
            )
            await self._handle_message(Message(reply_envlp, reply_msg))
        except Exception as e:
            self.logger.error(
                "Missing information for sending failure message", exc_info=e
            )


class MessageTransportProtocol(metaclass=abc.ABCMeta):
    """
    Message Transfer Protocol as defined in fipa.

    According to fipa, MTP is used to carry out the physical transfer of messages between two ACCs.
    This is a base interface for all MTPs.
    See http://fipa.org/specs/fipa00067/SC00067F.html
    """

    def __init__(self) -> None:
        """
        Default constructor which declares an attribute named `acc`.

        This attribute should be set by the :class:`AgentCommunicationChannel` at registration time.
        """
        self.acc: None | AgentCommunicationChannel = None

    @abc.abstractmethod
    async def send(self, message: Message, address: str) -> bool:
        """
        Send the provided message to the provided address.

        If the sending operation fails, the coroutine will return ``False``.
        :param message: the message to send.
        :param address: where to send the message
        """
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def scheme(self) -> str:
        """Get the protocol's scheme this MTP is providing."""
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def address(self) -> str:
        """Get the address of the ACC when using this MTP."""
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def type(self) -> str:
        """
        Get this MTP's type.

        For example: fipa.mts.mtp.http.std
        """
        raise NotImplementedError()

    @abc.abstractmethod
    async def start(self) -> None:
        """
        Start the MTP.

        It must not block the caller, so the MTP must be started on its own thread or as
        a new task.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    async def stop(self) -> None:
        """Stop the MTP."""
        raise NotImplementedError()


class MessageContext(NamedTuple):
    """Data class containing the context of a message: url of the sender and MTP used."""

    from_: str
    via: MessageTransportProtocol



# coding: utf-8
"""
A simple module that demonstrate how to create a simple agent and a simple behavior.

This module defines:

 * An agent, called CustomAgent
 * One behavior, called HelloWorldBehavior
 * It also show how to launch the platform.
"""
import piaf.agent as agent

from piaf.behavior import Behavior


class HelloWorldBehavior(Behavior):
    """A behavior that uses the agent's logger to display an Hello World message."""

    async def action(self):
        """Body of the Behavior."""
        self.agent.logger.info("Hello world from HelloWorldBehavior !")


class CustomAgent(agent.Agent):
    """A simple agent using the the :class:`HelloWorldBehavior`."""

    def __init__(self, aid, platform):
        """Create a new agent and add the behavior."""
        super().__init__(aid, platform)

        # Create an instance of HelloWorldBehavior and add it to the agent
        b = HelloWorldBehavior(self)
        self.add_behavior(b)


if __name__ == "__main__":
    import logging
    from piaf.launcher import PlatformLauncher, AgentDescription

    # Configure logging level and handler to see things
    logging.getLogger().setLevel(logging.INFO)
    logging.getLogger().addHandler(logging.StreamHandler())

    # Create the platform launcher
    launcher = PlatformLauncher("localhost")

    # Now we can add our agent
    launcher.add_agent(AgentDescription("hello", CustomAgent))

    # The program will run until you hit Ctrl+C
    launcher.run()



# coding: utf-8
"""All exceptions and errors that can occur in the WebAPI."""
from __future__ import annotations

from typing import Any, Dict

from fastapi import HTTPException


class BadRequestException(HTTPException):
    """Base exception for all client-related errors (4XX)."""

    def __init__(
        self, code: int = 400, detail: Any = None, headers: Dict[str, Any] | None = None
    ) -> None:
        """
        Create a new :class:`BadRequestException` instance.

        :param code: HTTP code, must be 4XX (default is 400)
        :param detail: optional detail about what happened
        :param headers: additional headers
        """
        super().__init__(code, detail, headers)


class InternalServerError(HTTPException):
    """Base exception for all server-related errors (5XX)."""

    def __init__(
        self, code: int = 500, detail: Any = None, headers: Dict[str, Any] | None = None
    ) -> None:
        """
        Create a new :class:`InternalServerError` instance.

        :param code: HTTP code, must be 5XX (default is 500)
        :param detail: optional detail about what happened
        :param headers: additional headers
        """
        super().__init__(500, detail, headers)


class DuplicatedPlatformNameError(BadRequestException):
    """Client error indicating the provided platform's name already exists."""

    def __init__(self, name: str) -> None:
        """
        Create a new :class:`DuplicatedPlatformNameError` instance.

        :param name: the duplicated name.
        """
        super().__init__(detail=f"Duplicated platform name {name}")


class UnknownPlatformError(BadRequestException):
    """Client error telling there is no platform named like asked."""

    def __init__(self, name: str) -> None:
        """
        Create a new :class:`UnknownPlatformError` instance.

        :param name: the provided platform's name
        """
        super().__init__(detail=f"Unknown platform with name {name}")


class RedisError(InternalServerError):
    """Base class for Redis-related errors."""

    pass


class RedisUnavailableError(RedisError):
    """A server error telling the Redis database can't be reached for some reason."""

    def __init__(self) -> None:
        """Create a new :class:`RedisUnavailableError` instance."""
        super().__init__(detail="Can't reach the Redis database.")


class RedisConnectionNotInitialized(RedisError):
    """A server error telling the connection is not initialized yet."""

    def __init__(self) -> None:
        """Create a new :class:`RedisConnectionNotInitialized` instance."""
        super().__init__(detail="No Redis connection initialized.")



# coding: utf-8
from __future__ import annotations

import asyncio
import contextlib
import copy
import enum
import logging
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    MutableMapping,
    MutableSequence,
    Optional,
    Sequence,
    Type,
)

import piaf.comm
import piaf.comm.mts
import piaf.exceptions as ex
from piaf.audit import AGENTS_PARENT_TOPIC, PLATFORM_TOPIC, Event, EventManager, Topic
from piaf.service import AMSService

if TYPE_CHECKING:
    import piaf.agent  # noqa

__all__ = [
    "AgentPlatformFacade",
    "AgentManager",
    "PlatformState",
    "AgentPlatform",
    "Extension",
]


class AgentPlatformFacade:
    """
    Facade limiting access to platform functionalities.

    This facade is provided to show the interface between user agents and the platform
    but should be used directly.

    If the agent is a service, then unknown attributes and methods at this level are forwarded to the decorated platform, meaning that services have a full access to all platform functionalities.
    """

    def __init__(self, platform: AgentPlatform, is_service: bool = False):
        """
        Create a new :class:`AgentPlatformFacade` backed by the provided platform.

        :param platform: the real platform behind this facade
        :param is_service: if set to `True`, this facade enables full platform access.
        """
        self._platform: AgentPlatform = platform
        self._is_service: bool = is_service

    async def suspend(self, agent: piaf.comm.AID) -> None:
        """
        Suspend an agent.

        Even if it is possible to suspend an other agent by providing its AID, it
        should not be done. Only suspend yourself !

        :param agent: the AID of the agent
        """
        await self._platform.agent_manager.suspend(agent)

    async def wait(self, agent: piaf.comm.AID) -> None:
        """
        Put an agent in WAITING state.

        Even if it is possible to put in waiting state an other agent by providing
        its AID, it should not be done.

        :param agent: the AID of the agent
        """
        await self._platform.agent_manager.wait(agent)

    async def quit(self, agent: piaf.comm.AID) -> None:
        """
        Make this agent quit.

        It will request the agent to die, but it provide the ability to shutdown
        properly. Again, use it with your agent's AID, not others.

        :param agent: the AID of the agent
        """
        await self._platform.agent_manager.quit(agent)

    def send(self, message: piaf.comm.Message) -> None:
        """
        Send the provided message.

        This method can raise several exceptions if there is any issue when sending the
        message.

        :param message: the message to send.
        """
        self._platform._acc.forward(message)

    def get_state(self, agent: piaf.comm.AID) -> piaf.agent.AgentState:
        """
        Get the state of the agent.

        Only use your agent's AID.

        :param agent: the agent's AID
        """
        return self._platform.agent_manager.get_state(agent)

    def get_state_condition(self, agent: piaf.comm.AID) -> asyncio.Condition:
        """
        Get a condition synchronization primitive based on the agent's sate.

        Some building blocks, like user agent or behaviors need to be notified when the
        state of the agent changed. This will get a shared condition on it. Only use
        with your agent's AID.

        :param agent: the agent's AID
        """
        return self._platform.agent_manager.get_state_condition(agent)

    @property
    def name(self) -> str:
        """Get this platform's name."""
        return self._platform.name

    @property
    def evt_manager(self) -> EventManager:
        """Get the platform's :class:`EventManager`."""
        return self._platform.evt_manager

    @property
    def extensions(self) -> Dict[str, Extension]:
        """Get available platform extensions."""
        if self._is_service:
            return self._platform.extensions
        return {
            name: ext
            for name, ext in self._platform.extensions.items()
            if ext.require_service
        }

    def __getattr__(self, name: str) -> Any:
        """
        If the agent is a service (ie requires full access to the platform), forward the call to the real platform.

        Otherwise raise an :class:`AttributeError` exception.
        """
        if self._is_service:
            return getattr(self._platform, name)
        raise AttributeError(name)


class _AgentContext:
    """
    Dataclass used to store an agent's context.

    Stored information is:

    * the agent itself
    * the agent's state
    * the task associated to the agent
    * the agent's state condition synchronization primitive

    """

    def __init__(self, agent: piaf.agent.Agent):
        self.agent = agent
        self.state = piaf.agent.AgentState.INITIATED
        self.task: Optional[asyncio.Future[None]] = None
        self.state_condition = asyncio.Condition()


class AgentManager:
    """
    Object responsible of managing agents on a platform.

    This manager is dedicated to agent management and is a part of the platform.
    It is responsible of the agent's creation and death and it handle state
    modifications.

    User agents shouldn't interact with it directly but rather through the
    :class:`AgentPlatformFacade`.
    """

    def __init__(
        self, platform: AgentPlatform, acc: piaf.comm.mts.AgentCommunicationChannel
    ):
        """
        Create a new manager associated to the provided platform.

        :param platform: the platform that will use this manager.
        :param acc: full access to the Agent Communication Channel
        """
        self._contexts: MutableMapping[piaf.comm.AID, _AgentContext] = {}
        self._platform = platform
        self._acc = acc
        self.logger = logging.getLogger(type(self).__name__)

    async def wait(self, agent: piaf.comm.AID) -> None:
        """
        Switch the agent's state to WAITING.

        :param agent: the agent's AID
        :raise ex.StateTransitionException: if the agent's state wasn't ACTIVE
        """
        await self._switch_state(
            agent, piaf.agent.AgentState.ACTIVE, piaf.agent.AgentState.WAITING
        )

    async def suspend(self, agent: piaf.comm.AID) -> None:
        """
        Switch the agent's state to SUSPENDED.

        :param agent: the agent's AID
        :raise ex.StateTransitionException: if the agent's state wasn't ACTIVE
        """
        await self._switch_state(
            agent, piaf.agent.AgentState.ACTIVE, piaf.agent.AgentState.SUSPENDED
        )

    async def resume(self, agent: piaf.comm.AID) -> None:
        """
        Switch the agent's state to ACTIVE.

        :param agent: the agent's AID
        :raise ex.StateTransitionException: if the agent's state wasn't SUSPENDED
        """
        await self._switch_state(
            agent, piaf.agent.AgentState.SUSPENDED, piaf.agent.AgentState.ACTIVE
        )

    async def wake_up(self, agent: piaf.comm.AID) -> None:
        """
        Switch the agent's state to ACTIVE.

        :param agent: the agent's AID
        :raise ex.StateTransitionException: if the agent's state wasn't WAITING
        """
        await self._switch_state(
            agent, piaf.agent.AgentState.WAITING, piaf.agent.AgentState.ACTIVE
        )

    async def create(
        self,
        agent_class: Type[piaf.agent.Agent],
        short_name: str,
        is_service: bool = False,
        *args,
        **kwargs,
    ) -> piaf.comm.AID:
        """
        Create a new agent.

        It will use the provided agent class and arguments to create a new agent.
        Agent's state will be set to INITIALIZED and the AID will be built according to
        the provided name and the platform parameters (name, supported schemes,
        AMS AID ...).

        :param agent_class: the agent class that is used to instantiate the agent
        :param short_name: the agent's name, without the platform name
        :param is_service: if set to True, will have full access to the platform. Use it with caution ! Default is
                           False.
        :param args: extra arguments used to instantiate the agent (other than the
                     platform)
        :param kwargs: same as ``args`` but for keywords arguments
        :raise DuplicatedNameException: if an agent with that name already exists
        :raise InvalidStateException: if the platform is not ready.
        """
        # assert self._platform.ams is not None
        if not self._platform.state == PlatformState.RUNNING:
            raise ex.InvalidStateException(self._platform.state, "add_agent")

        # Create a temporary AID
        aid = piaf.comm.AID(f"{short_name}@{self._platform.name}")
        if aid in self._contexts:
            raise ex.DuplicatedNameException(short_name)

        aid.addresses = self._acc.addresses

        # Create agent
        agent = agent_class(
            aid, AgentPlatformFacade(self._platform, is_service), *args, **kwargs
        )
        self._contexts[aid] = _AgentContext(agent)

        # Register agent to the ACC
        self._acc.register_agent_or_service(short_name, agent.mailbox)  # type: ignore

        # Fire event
        await self._platform.evt_manager.publish(
            Event(aid.name, "agent_creation", aid),
            [
                Topic.resolve_from_parent_topic(AGENTS_PARENT_TOPIC, aid.short_name),
                AGENTS_PARENT_TOPIC,
            ],
        )

        return aid

    async def stop_all(self, timeout: float = 5.0) -> None:
        """
        Stop all running agents.

        It will first try to do a nice stop for each agent and then, if some are taking
        to much time, it will terminate the remaining.

        :param timeout: how much time is given to shutdown properly all agents (default
                        is 5s)
        """
        if not self._contexts:
            return

        tasks = [
            context.task
            for context in self._contexts.values()
            if context.task is not None
        ]

        for context in self._contexts:
            await self.quit(context)

        _, s_pending = await asyncio.wait(tasks, timeout=timeout)
        pending = list(s_pending)

        for task in pending:
            self.logger.warning(
                "[AgentManager] Agent %s did not terminate in time, "
                "emergency stop requested.",
                task.name,  # type: ignore
            )
            task.cancel()

        # No need to await termination, display exceptions and cleanup contexts
        # It is done in _clean_up task.

    async def invoke(self, agent: piaf.comm.AID) -> None:
        """
        Invoke a previously created agent.

        The agent's state will switch from INITIATED to ACTIVE and the run method will
        be scheduled.

        :param agent: the agent that will be invoked
        """
        await self._switch_state(
            agent, piaf.agent.AgentState.INITIATED, piaf.agent.AgentState.ACTIVE
        )
        self._contexts[agent].task = asyncio.create_task(
            self._contexts[agent].agent.run()
        )
        self._contexts[agent].task.name = agent.short_name  # type: ignore
        self.logger.info(f"Please welcome Agent {agent.short_name} !")

        # Schedule the cleanup task
        asyncio.create_task(self._clean_up(agent))

    async def quit(self, agent: piaf.comm.AID) -> None:
        """
        Instruct the agent that it should die.

        This will initiate a "graceful" stop. The agent state will switch to UNKNOWN
        (as requested by fipa). This doesn't await the agent termination, as this method
        can be called from the agent itself.
        """
        if agent not in self._contexts:
            raise ex.IllegalArgumentException(f"Unknown agent: {agent}")

        self._contexts[agent].state = piaf.agent.AgentState.UNKNOWN
        async with self._contexts[agent].state_condition:
            self._contexts[agent].state_condition.notify_all()

    async def _clean_up(self, agent: piaf.comm.AID) -> None:
        """
        Given an agent, wait its termination and removed it from the manager.

        :param agent: the agent to remove
        """
        task = self._contexts[agent].task
        assert task is not None
        result = (await asyncio.gather(task, return_exceptions=True))[0]

        if result is not None:
            self.logger.exception(
                "[AgentManager] Agent %s raised an exception.",
                task.name,  # type: ignore
                exc_info=result,
            )

        del self._contexts[agent]
        self._acc.forget_agent_or_service(agent.short_name)

        await self._platform.evt_manager.publish(
            Event(agent.name, "agent_death", None),
            [
                Topic.resolve_from_parent_topic(AGENTS_PARENT_TOPIC, agent.short_name),
                AGENTS_PARENT_TOPIC,
            ],
        )

    async def terminate(self, agent: piaf.comm.AID) -> None:
        """
        Terminate an agent (forced).

        This will cancel the agent task and then clean up the manager. Unlike
        :meth:`stop`, this method waits the agent's death. Try to use this method only
        when the :meth:`stop` method is not working.

        :param agent: the agent to terminate
        """
        task = self._contexts[agent].task
        if task is not None:
            task.cancel()

    def get_state(self, agent: piaf.comm.AID) -> piaf.agent.AgentState:
        """
        Get the state of the provided agent.

        If the agent doesn't exist or is dying/dead, this method return UNKNOWN.

        :param agent: the agent
        """
        result = piaf.agent.AgentState.UNKNOWN
        with contextlib.suppress(KeyError):
            result = self._contexts[agent].state
        return result

    def get_state_condition(self, agent: piaf.comm.AID) -> asyncio.Condition:
        """
        Get the state synchronization primitive associated to the agent.

        :param agent: the agent you want the synchronization primitive
        """
        return self._contexts[agent].state_condition

    async def _switch_state(
        self,
        agent: piaf.comm.AID,
        from_: piaf.agent.AgentState,
        to: piaf.agent.AgentState,
    ) -> None:
        """
        Transition an agent's state to another.

        It cares about the current agent's state and will raise an exception if it is
        not the one expected.

        :param agent: the agent
        :param from_: what should be the current agent's state
        :param to_: the agent state you want
        :raise ex.StateTransitionException: the expected agent's state is not the
                                            current agent's state
        :raise ex.IllegalArgumentException: the agent doesn't exist (or is dead)
        """
        if agent not in self._contexts:
            raise ex.IllegalArgumentException(f"Unknown agent: {agent}")

        if not self._contexts[agent].state == from_:
            raise ex.StateTransitionException(self._contexts[agent].state, to)

        self._contexts[agent].state = to
        async with self._contexts[agent].state_condition:
            self._contexts[agent].state_condition.notify_all()
        await self._platform.evt_manager.publish(
            Event(agent.name, "state_change", {"from": from_, "to": to}),
            Topic.resolve_from_parent_topic(AGENTS_PARENT_TOPIC, agent.short_name),
        )
        self.logger.debug(
            "Agent %s switched its state from %s to %s", agent.name, from_, to
        )

    def get_agents(
        self, state: piaf.agent.AgentState | None = None
    ) -> Sequence[piaf.comm.AID]:
        """
        Get all known agents.

        If `state` is provided, the result will only contain agents having the given state.

        :param state: If provided, filter result and keep only agents having the given state.
        """
        agents = self._contexts.keys()
        if state is not None:
            agents = (agt for agt in agents if self.get_state(agt) == state)
        return [copy.deepcopy(agent) for agent in agents]


class Extension:
    """
    Extensions are additional functionalities loaded by the platform before anything else.

    They can be made available to all agents or only services (default). This interface defines the bare minimum any extension should implement.
    """

    def __init__(self) -> None:
        """Create a new :class:`Extension` instance."""
        self.platform: None | AgentPlatform = None

    async def on_start(self):
        """Do startup things, default nothing."""
        pass

    async def on_stop(self):
        """Do closing things, default nothing."""
        pass

    @property
    def require_service(self):
        """Tell if the extension is reserved for services (default) or not."""
        return False


class PlatformState(enum.Enum):
    """An enum that represents what can be a platform's state."""

    def _generate_next_value_(name, start, count, last_values):  # noqa
        return name

    INITIALIZED = enum.auto()
    RUNNING = enum.auto()
    STOPPED = enum.auto()


class AgentPlatform:
    """
    The :class:`AgentPlatform` is the object you need to run your agents.

    The first need to do is to instantiate a new platform. Them you must configure it.
    For now, the only required thing is to specify the supported schemes. Schemes are
    used for the communication between agents and is often a protocol's name (like http,
    ftp ...).
    There is one built-in scheme called `memory`.

    The next thing to do is to start the platform, using the :meth:`start` method. This
    is to ensure that the transport system and the AMS are up and ready for your agents.

    The last thing to do is to create and invoke your agents. Use the
    :class:`AgentManager` bounded to the platform to do it.

    Here is an example of how to create a platform and to run some agents::

        import asyncio
        import piaf.ptf
        import piaf.comm
        import piaf.comm.mts
        import your.agent

        # Configure logging level and handler to see debug information
        logging.getLogger().setLevel(logging.DEBUG)
        logging.getLogger().addHandler(logging.StreamHandler())

        # Create agent platform
        ap = piaf.ptf.AgentPlatform("localhost")

        async def main():
            '''Coroutine that starts the platform and add agents.'''
            # Before adding our agents, we need to start the platform. This ensure that
            # the AMS agent is created
            await ap.start()

            # We could add extension or MTP
            # await ap.acc.register_mtp(...)

            # Now we can add our agent(s)
            aid = await ap.agent_manager.create(your.agent.First, "F")
            await ap.agent_manager.invoke(aid)

            aid = await ap.agent_manager.create(your.agent.First, "S")
            await ap.agent_manager.invoke(aid)

        # We are using asyncio library to run our example
        # The program will run until you hit Ctrl+C
        loop = asyncio.get_event_loop()
        loop.set_debug(True)
        try:
            loop.create_task(main())
            loop.run_forever()
        except KeyboardInterrupt:
            pass
        finally:
            loop.run_until_complete(ap.stop())
            loop.run_until_complete(loop.shutdown_asyncgens())
            loop.close()

    """

    class ACCDelegate:
        """
        Delegate object that provide mtp registration.

        Real object (delegate) is the :class:`piaf.comm.mts.AgentCommunicationChannel` instance.
        """

        def __init__(self, acc: piaf.comm.mts.AgentCommunicationChannel) -> None:
            """Create delegate with the provided real ACC."""
            self._delegate = acc

        async def register_mtp(
            self, mtp: piaf.comm.mts.MessageTransportProtocol
        ) -> None:
            """
            Register a new MTP and start it.

            Once registered, the ACC will be able to use the MTP to send messages. ACC doesn't allow duplicated MTPs
            (ie, MTP providing the same scheme). If you try to register a MTP with an already registered scheme, this
            method will raise an exception.

            :raise DuplicatedSchemeException: if you try to register an MTP handling an already-register scheme.
            """
            await self._delegate.register_mtp(mtp)

    def __init__(self, name: str, evt_manager: None | EventManager = None):
        """
        Create a new :class:`AgentPlatform`. Initial state is ``INITIALIZED``.

        :param name: this platform's name.
        """
        self.logger = logging.getLogger(type(self).__name__)
        self._evt_manager = evt_manager if evt_manager is not None else EventManager()
        self._name = name
        self._acc = piaf.comm.mts.AgentCommunicationChannel(self)
        self._agt_manager = AgentManager(self, self._acc)
        self._tasks: MutableSequence[asyncio.Future[Optional[Any]]] = []
        self._state = PlatformState.INITIALIZED
        self._state_sync = asyncio.Condition()
        self._extensions: Dict[str, Extension] = {}
        self.logger.info(f"Platform {self.name} initialized")

    def load_extension(self, name: str, extension: Extension) -> None:
        """
        Load the provided extension under the given name.

        :param name: some name for indexing
        :param extension: the extension to load
        :raise InvalidStateException: Platform is not in the INITIALIZED state
        """
        if self.state != PlatformState.INITIALIZED:
            raise ex.InvalidStateException(self.state, "load_extension")
        self._extensions[name] = extension
        extension.platform = self
        self.logger.info(f"Loaded extension {name}")

    @property
    def name(self) -> str:
        """Get this platform's name."""
        return self._name

    @property
    def schemes(self) -> Sequence[str]:
        """Get his platform supported schemes."""
        return self._acc.schemes

    @property
    def state(self) -> PlatformState:
        """Get this platform's state."""
        return self._state

    @property
    def state_sync(self) -> asyncio.Condition:
        """A synchronization primitive on the platform's state."""
        return self._state_sync

    @property
    def agent_manager(self) -> AgentManager:
        """Get the agent manager associated to this platform."""
        return self._agt_manager

    @property
    def acc(self) -> AgentPlatform.ACCDelegate:
        """Get limited access to the AgentCommunicationChannel of the platform."""
        return AgentPlatform.ACCDelegate(self._acc)

    @property
    def evt_manager(self):
        """Get the unique :class:`piaf.events.EventManager` instance."""
        return self._evt_manager

    @property
    def extensions(self) -> Dict[str, Extension]:
        """Get all loaded extensions."""
        return self._extensions

    async def start(self) -> None:
        """
        Start the platform.

        The platform state will transition to ``RUNNING`` and the AMS agent will be
        started.

        :raise StateTransitionException: the platform's state is different from `INITIALIZED`
        """
        if self._state != PlatformState.INITIALIZED:
            raise ex.StateTransitionException(self._state, PlatformState.RUNNING)

        # Set platform state
        async with self.state_sync:
            self._state = PlatformState.RUNNING
            self.state_sync.notify_all()

        # Load extensions
        for _, ext in self.extensions.items():
            await ext.on_start()

        await self.evt_manager.publish(
            Event(
                "platform",
                "state_change",
                {"from": PlatformState.INITIALIZED, "to": PlatformState.RUNNING},
            ),
            PLATFORM_TOPIC,
        )
        self.logger.info("Platform is running.")

        # Create AMS service
        ams_aid = await self._agt_manager.create(AMSService, "ams", True)
        await self._agt_manager.invoke(ams_aid)

    async def stop(self) -> None:
        """
        Stop this platform.

        It will stop all running agents (make them quit) and the platform's state will
        be ``STOPPED``.

        :raise StateTransitionException: the platform's state is different from `RUNNING`
        """
        if self._state != PlatformState.RUNNING:
            raise ex.StateTransitionException(self._state, PlatformState.STOPPED)

        await self._agt_manager.stop_all()
        await self._acc.stop()

        # Stop extensions
        for _, ext in self.extensions.items():
            await ext.on_stop()

        if self._tasks:
            for task in self._tasks:
                task.cancel()
            await asyncio.wait(self._tasks)

        async with self.state_sync:
            self._state = PlatformState.STOPPED
            self.state_sync.notify_all()

        # Schedule the publishing of the event rather than waiting it.
        # This is because some cleanup functions can be triggered by the state change
        # and close the loop before the event is published
        self.evt_manager.publish_sync(
            Event(
                "platform",
                "state_change",
                {"from": PlatformState.RUNNING, "to": PlatformState.STOPPED},
            ),
            PLATFORM_TOPIC,
        )
        self.logger.info(f"Platform {self.name} has been stopped.")



# coding: utf-8
"""Define all exceptions that are used in the piaf framework."""
from __future__ import annotations
from typing import Union, TYPE_CHECKING

if TYPE_CHECKING:
    import piaf.agent  # noqa
    import piaf.ptf as plt  # noqa

__all__ = [
    "InvalidStateException",
    "StateTransitionException",
    "UnsupportedOperationException",
    "MessageNotSentException",
    "MandatoryFieldValueException",
    "DuplicatedNameException",
    "DuplicatedSchemeException",
    "IllegalArgumentException",
]


class InvalidStateException(Exception):
    """
    Exception raised when an operation cannot be performed because of an invalid state.

    The exception message contains the state and the operation that was attempted.
    """

    def __init__(
        self,
        state: piaf.agent.AgentState | plt.PlatformState,
        operation: str = "",
    ):
        """
        Create a new :class:`InvalidStateException`.

        :param state: object's state when the exception occurred.
        :param operation: the operation that was attempted when the exception occurred.
        """
        super().__init__(f"Cannot perform operation {operation} when state is {state}.")


class StateTransitionException(Exception):
    """
    Exception raised when the object can't transition from a state to another.

    The exception's message contains the original state and the state that was targeted.
    """

    def __init__(
        self,
        from_: piaf.agent.AgentState | plt.PlatformState,
        to: piaf.agent.AgentState | plt.PlatformState,
    ):
        """
        Create a new :class:`StateTransitionException`.

        :param from_: the object's state
        :param to_: the state that can't be reached
        """
        super().__init__(f"Cannot transition from state {from_} to state {to}.")


class UnsupportedOperationException(Exception):
    """
    Exception raised when the requested operation is not supported.

    The message can provide extra information about the reasons.
    """

    def __init__(self, msg: str = ""):
        """
        Create a new :class:`UnsupportedOperationException`.

        :param msg: why this operation is not supported (optional)
        """
        super().__init__(msg)


class MessageNotSentException(Exception):
    """Raised when something went wrong and a message sending failed."""

    def __init__(self) -> None:
        """Create a new :class:`MessageNotSentException`."""
        super().__init__(
            "Your message couldn't be sent, see MTS logs for more information."
        )


class MandatoryFieldValueException(Exception):
    """Raised when a mandatory field as no value."""

    def __init__(self, field_name) -> None:
        """Create a new :class:`MandatoryFieldValueException`."""
        super().__init__(f"Mandatory field '{field_name}' has no value.'")


class DuplicatedNameException(Exception):
    """Raised when some action doesn't support name duplication."""

    def __init__(self, name: str) -> None:
        """
        Create a new :class:`DuplicatedNameException`.

        :param name: duplicated name
        """
        super().__init__(f"Duplicated name: {name}")


class DuplicatedSchemeException(Exception):
    """Raised when you try to register a MTP into an ACC for a scheme that as already a registered MTP."""

    def __init__(self, scheme: str) -> None:
        """
        Create a new :class:`DuplicatedSchemeException`.

        :param scheme: duplicated scheme
        """
        super().__init__(f"Duplicated scheme: {scheme}")


class IllegalArgumentException(Exception):
    """Raised when an argument doesn't respect a precondition."""

    def __init__(self, message: str) -> None:
        """
        Create a new :class:`IllegalArgumentException`.

        :param message: message to display. Mostlikely the violated precondition.
        """
        super().__init__(message)



"""
Part of the "two-platforms" test.

This module contains everything to setup and run the second platform. Once invoked, it will
create a platform named "ptf2" and one agent, called "other".

The agent waits until a message is received, print it using its logger and die.

.. note: In order to run this example, you will have to setup an AMQP server. Using Docker and RabbitMQ,
    you can easily launch the example: `docker run -it --rm --name rabbitmq -p 5672:5672 -p 15672:15672 rabbitmq:3.9-management-alpine`

Once the server is ready, just run `python ptf2.py` in one terminal and `python ptf1.py` in another. If the example
works, you should see in the ptf2 terminal the message sent by the ptf1 agent.
"""
from piaf.agent import Agent
from piaf.behavior import Behavior, FSMBehavior, SuicideBehavior
from piaf.comm import AID
from piaf.ptf import AgentPlatformFacade


class RcvMsgBehavior(Behavior):
    """A simple behavior that receives messages and print them using the agent's logger."""

    async def action(self):
        """Receive a message and display it."""
        msg = await self.agent.receive()
        self.agent.logger.info(
            "[%s] Received %s", self.agent.aid.short_name, msg.acl_message.content
        )


class OtherAgent(Agent):
    """A simple agent that waits a message, print it and then die."""

    def __init__(self, aid: AID, platform: AgentPlatformFacade):
        """
        Initialize a new instance of the agent.

        :param aid: the agent's AID
        :param platform: where the agent will run.
        """
        super().__init__(aid, platform)

        fsm = FSMBehavior(self)
        self.add_behavior(fsm)

        fsm.add_state("RCV", RcvMsgBehavior)
        fsm.add_state("DEATH", SuicideBehavior, final=True)
        fsm.set_initial_state("RCV")

        fsm.add_transition("RCV", "DEATH", lambda e: True)


if __name__ == "__main__":
    import logging
    from piaf.launcher import AgentDescription, PlatformLauncher
    from piaf.comm.mtp.amqp import AMQPMessageTransportProtocol

    # Configure logging level and handler to see things
    logging.getLogger().setLevel(logging.INFO)
    logging.getLogger().addHandler(logging.StreamHandler())

    # Create platform launcher
    launcher = PlatformLauncher("ptf2")

    # Since we want to connect this platform to another one, we need an MTP.
    # Here we are going to use the AMQPMessageTransferProtocol, which relies on AMQP 0.9.1 protocol.
    launcher.add_mtp(AMQPMessageTransportProtocol("amqp://guest:guest@localhost/"))

    # Now we can add our agent
    launcher.add_agent(AgentDescription("other", OtherAgent))

    # The program will run until you hit Ctrl+C
    launcher.run()



# coding: utf-8
"""
A small program testing the DF capabilities.

This module defines a service called :class:`FIPAWebService` which can send HTML pages from the `FIPA Website(http://fipa.org)`_
"""
from __future__ import annotations
import asyncio
from http.client import HTTPConnection, HTTPResponse

from piaf.agent import Agent
from piaf.behavior import Behavior
from piaf.service import DFAgentDescription, DFService, ServiceDescription
from piaf.comm import AID, MT_CONVERSATION_ID, ACLMessage, Performative
from piaf.ptf import AgentPlatformFacade
from piaf.util import (
    FIPARequestProtocolBehavior,
    agree_message_from_request,
    inform_message_from_request,
    not_understood_message_from_request,
)


class FIPAWebPageService(Agent):
    """
    A service which sends the first 200 characters of FIPA website page.

    The service supports the FIPA Request protocol. The message content is expected to be the page path, in str format.
    """

    def __init__(self, aid: AID, platform: AgentPlatformFacade):
        super().__init__(aid, platform)

        # Initialize behaviors
        self.add_behavior(FIPAWebPageRequestProtocolBehavior(self))
        self.add_behavior(RegisterFIPAWebPageServiceBehavior(self))


class FIPAWebPageRequestProtocolBehavior(FIPARequestProtocolBehavior):
    """
    This behavior handle incoming request and retrieves the requested page.

    It supports the full Request Protocol and sends an Agree message before sending the Inform message.
    """

    def __init__(self, agent: Agent):
        """
        Initialize the behavior.

        It will create an HTTP connection to the FIPA website.
        """
        super().__init__(agent)

        self.connection: HTTPConnection = HTTPConnection("fipa.org")
        self.lock = asyncio.Lock()

    def check_message(self, msg: ACLMessage) -> bool:
        """
        Check if the given message is valid, i.e. the agent understands the content.

        If not, then replies with a NOT_UNDERSTOOD message with a reason.

        :param msg: message to check
        :return: `True` if the message is valid, `False` otherwise.
        """
        if not isinstance(msg.content, str):
            self.agent.send(
                not_understood_message_from_request(
                    msg, "Content is expected to be an str value."
                )
            )
            return False
        return True

    async def on_valid_request(self, msg: ACLMessage) -> None:
        """
        Behavior executed once we are sure the received message is valid.

        First it sends an AGREE message and then it uses an internal :class:`HTTPConnection` to request the
        the full page at the given path.

        :param msg: the valid message
        """
        self.agent.send(agree_message_from_request(msg))
        async with self.lock:
            response = self._retrieve_page(msg.content)
            self.agent.send(inform_message_from_request(msg, response.read(200)))
            response.read()  # Read all so we can re-use the connection for the next time. May take time.

    def _retrieve_page(self, url: str) -> HTTPResponse:
        """
        Given a relative path, use the internal connection to the FIPA website to issue a GET request and return the received response.

        :param url: page path, starting with '/'.
        """
        self.connection.request("GET", url)
        return self.connection.getresponse()


class RegisterFIPAWebPageServiceBehavior(Behavior):
    """
    A one-shot behavior which registers our :class:`FIPAWebPageService` to the DF agent.
    """

    async def action(self) -> None:
        register_msg = (
            ACLMessage.Builder()
            .performative(Performative.REQUEST)
            .conversation_id("fipa-web-page-service")
            .receiver(AID(f"DF@{self.agent.aid.hap_name}"))
            .content(
                [
                    DFService.REGISTER_FUNC,
                    DFAgentDescription(
                        self.agent.aid,
                        services=(
                            ServiceDescription(
                                "Fipa web page service",
                                protocols=("fipa-request",),
                            ),
                        ),
                        protocols={"fipa-request"},
                    ),
                ]
            )
            .build()
        )

        self.agent.send(register_msg)


class TestingAgent(Agent):
    """
    A testing agent requesting the FIPA00023 specification.
    """

    def __init__(self, aid: AID, platform: AgentPlatformFacade):
        super().__init__(aid, platform)

        self.add_behavior(AskFIPA00023Behavior(self))


class AskFIPA00023Behavior(Behavior):
    """
    Request the FIPA00023 specification page to the first FIPA web service found.

    It uses the DF agent to search an agent providing the service and if such agent exists, it then sends a
    request to retrieve the page.
    """

    PAGE = "/specs/fipa00023/SC00023K.html"
    FIND_CONV = "find-fipa-web-page-service"
    ASK_CONV = "ask-page-fipa00023"

    def __init__(self, agent: Agent):
        super().__init__(agent)
        self.is_done = False

    def done(self) -> bool:
        return self.is_done

    async def action(self) -> None:
        # Find FIPAWebPageService
        find = (
            ACLMessage.Builder()
            .performative(Performative.REQUEST)
            .conversation_id(self.FIND_CONV)
            .receiver(AID(f"DF@{self.agent.aid.hap_name}"))
            .content(
                [
                    DFService.SEARCH_FUNC,
                    DFAgentDescription(
                        services=(
                            ServiceDescription(
                                "Fipa web page service",
                                protocols=("fipa-request",),
                            ),
                        ),
                    ),
                ]
            )
            .build()
        )
        self.agent.send(find)

        # Wait until we get the response
        _ = await self.agent.receive(MT_CONVERSATION_ID(self.FIND_CONV))
        response = await self.agent.receive(MT_CONVERSATION_ID(self.FIND_CONV))

        # Extract service aid
        # According to DF spect, content is a tuple (request content, reponse)
        # The response is a list of matching agents, we take the first one
        if len(response.acl_message.content[1]) == 0:
            self.agent.logger.info("Service not available yet.")
            return
        service_aid = response.acl_message.content[1][0].name
        self.is_done = True

        # Contact service and ask page
        ask = (
            ACLMessage.Builder()
            .performative(Performative.REQUEST)
            .conversation_id(self.ASK_CONV)
            .receiver(service_aid)
            .content(self.PAGE)
            .build()
        )
        self.agent.send(ask)

        # Wait until we get the response
        _ = await self.agent.receive(MT_CONVERSATION_ID(self.ASK_CONV))
        response = await self.agent.receive(MT_CONVERSATION_ID(self.ASK_CONV))

        # Display the first 200 chars
        self.agent.logger.info(
            f"[{self.agent.aid.short_name}]: {response.acl_message.content}"
        )


if __name__ == "__main__":
    import logging
    import piaf.launcher as pl

    # Configure logging level and handler to see things
    logging.getLogger().setLevel(logging.INFO)
    logging.getLogger().addHandler(logging.StreamHandler())

    # Create the platform launcher
    launcher = pl.PlatformLauncher("localhost")

    # Add both DF FIPA services
    launcher.add_service(pl.ServiceDescription("DF", DFService))
    launcher.add_service(pl.ServiceDescription("FIPAService", FIPAWebPageService))

    # Add testing agents
    for i in range(10):
        launcher.add_agent(pl.AgentDescription(f"testing-{i}", TestingAgent))

    # The program will run until you hit Ctrl+C
    launcher.run()



# coding: utf-8
"""
A small program testing the AMS capabilities.

The scenario is the following:

1. Ten receiving agents and one broadcaster are created
2. The broadcaster one perform a request to find active agents
3. It sends a `Hello World` message to inform agents of its presence
4. The receiving agents display the message

It is possible that not all agents receives the message, since the broadcaster dies just after having sent its message
to at least one active agent. So, if some receivers where not ready at the time the broadcaster queried the AMS, those
agents won't receive the message.
"""
from __future__ import annotations
import asyncio
from typing import TYPE_CHECKING, List

from piaf.agent import Agent, AgentState
from piaf.service import AMSService, AMSAgentDescription
from piaf.comm import (
    MT_CONVERSATION_ID,
    ACLMessage,
    Message,
    MessageTemplate,
    Performative,
    AID,
)
from piaf.behavior import Behavior

if TYPE_CHECKING:
    from piaf.ptf import AgentPlatform, AgentPlatformFacade


class SendHello(Behavior):
    """
    One shot behavior to say hello to all active agents.

    Use the AMS to search active agents and send an `INFORM` message. The behavior is over when at least one agent is
    returned by the AMS.
    """

    def __init__(self, agent: Agent):
        super().__init__(agent)
        self.is_done = False

    def done(self) -> bool:
        return self.is_done

    async def action(self) -> None:
        """
        Behavior `action` method.

        1. Contact AMS and ask all active agents
        2. If AMS respond with INFORM
            1. Loop over received :class:`AMSAgentDescription` objects
            2. If neither owner nor the AMS, send hello

        """
        # First, create a message to query the AMS
        msg = (
            ACLMessage.Builder()
            .performative(Performative.REQUEST)
            .receiver(AID(f"ams@{self.agent.aid.hap_name}"))
            .conversation_id(f"{self.agent.aid.name}-sendhello")
            .content(
                [
                    AMSService.SEARCH_FUNC,
                    AMSAgentDescription(None, None, AgentState.ACTIVE),
                ]
            )
            .build()
        )
        self.agent.send(msg)

        # Wait a reply. First it should receive an AGREE, then an INFORM with the query response
        mt: MessageTemplate = MT_CONVERSATION_ID(msg.conversation_id)
        reply: Message = await self.agent.receive(mt)
        if reply.acl_message.performative != Performative.AGREE:
            return

        # If it succeeded, loop over content
        reply = await self.agent.receive(mt)
        if reply.acl_message.performative == Performative.INFORM:

            # Extract content
            content: List[AMSAgentDescription] = reply.acl_message.content[1]

            # No remaining agent, try again later
            if not content:
                asyncio.sleep(1)
                return

            # At least one active agent is a receiver
            for agt_description in content:
                other_aid = agt_description.name

                # If the description is neither agent nor the AMS, send hello !
                if other_aid != self.agent.aid and other_aid.short_name != "ams":
                    self.agent.send(
                        ACLMessage.Builder()
                        .performative(Performative.INFORM)
                        .receiver(other_aid)
                        .conversation_id(
                            f"{self.agent.aid.name}-sendhello-{other_aid.name}"
                        )
                        .content("Hello friend!")
                        .build()
                    )
                    self.is_done = True


class DisplayReceivedMessage(Behavior):
    """One shot behavior displaying the first received message."""

    async def action(self) -> None:
        """Use the agent's logger to display the first received message."""
        msg = await self.agent.receive()
        self.agent.logger.info(
            f"From {msg.acl_message.sender}: {msg.acl_message.content}"
        )


class BroadcastAgent(Agent):
    """
    Simple agent broadcasting a 'Hello' message to other active agents.

    The broadcast is performed when the agent is invoked. Agent dies immediately after.
    """

    def __init__(
        self,
        aid: AID,
        platform: AgentPlatformFacade | AgentPlatform,
    ):
        super().__init__(aid, platform)

        self.add_behavior(SendHello(self))


class DisplayAgent(Agent):
    """A simple agent displaying the first received message and dying right after that."""

    def __init__(self, aid: AID, platform: AgentPlatformFacade | AgentPlatform):
        super().__init__(aid, platform)

        self.add_behavior(DisplayReceivedMessage(self))


if __name__ == "__main__":
    import logging
    from piaf.launcher import AgentDescription, PlatformLauncher

    # Configure logging level and handler to see things
    logging.getLogger().setLevel(logging.INFO)
    logging.getLogger().addHandler(logging.StreamHandler())

    # Create the platform launcher
    launcher = PlatformLauncher("localhost")

    # Add agents
    for i in range(10):
        launcher.add_agent(AgentDescription(f"DA-{i}", DisplayAgent))
    launcher.add_agent(AgentDescription("BA", BroadcastAgent))

    # The program will run until you hit Ctrl+C
    launcher.run()



# coding: utf-8
"""A collection of predefined tasks to interact from the WebAPI with a simulation."""
from __future__ import annotations

import abc
import asyncio
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Dict, Iterable, List, Tuple
from uuid import UUID, uuid4

from piaf.agent import AgentState
from piaf.comm import AID, MT_CONVERSATION_ID, ACLMessage, Performative
from piaf.service import AgentCreationDescription, AMSAgentDescription, AMSService

if TYPE_CHECKING:
    from piaf.agent import Agent
    from piaf.api.impl.process import APIAgent


class Task(metaclass=abc.ABCMeta):
    """
    An abstraction of a task.

    Each task gets a unique ID number so a response can later be associated. Concrete class should implement two methods:

    - :meth:`Task.from_json` which creates an instance from a JSON-like structure
    - :meth:`Task.execute` which does the actual work
    """

    @staticmethod
    @abc.abstractmethod
    def from_json(json_repr: Dict[str, Any]) -> Task:
        """
        Unpack the given JSON object into a :class:`Task`.

        :param json_repr: the JSON object
        :return: a :class:`Task` object or one of its subclasses
        """
        raise NotImplementedError()

    def __init__(self) -> None:
        """
        Initialize a new :class:`Task` object.

        It generates a new UUID4 an set it as the task's ID.
        """
        self._id: UUID = uuid4()

    @property
    def id(self) -> UUID:
        """
        Get this task's ID.

        :return: an UUID
        """
        return self._id

    @abc.abstractmethod
    async def execute(self, agent: APIAgent) -> Any:
        """
        Realize the work the task is supposed to do.

        :param agent: the :class:`APIAgent` that is executing the task
        :return: whatever the task returns
        """
        raise NotImplementedError()


@dataclass(frozen=True, eq=True)
class TaskResult:
    """
    A small dataclass that stores a task's execution result.

    This class has three fields:

    - error: stores an exception if the execution failed
    - result: the task's execution result, if any
    - task_id: the task's id this result is linked to

    """

    error: BaseException | None
    result: Any
    task_id: UUID


class CreateAgentTask(Task):
    """
    A task that creates a new agent in the platform.

    Given an :class:`piaf.service.AgentCreationDescription` serialized in a JSON object, this task asks the AMS to create and initialize an agent.

    On a successful execution, the task returns the AID of the created agent.
    """

    @staticmethod
    def from_json(json_repr: Dict[str, Any]) -> CreateAgentTask:
        """
        Create a new :class:`CreateAgentTask` from the given JSON object.

        The object must have the following fields:

        - class_name: qualified name of the agent's class
        - agent_name: short name of the agent
        - args: extra arguments given to the agent's constructor
        - is_service: tell if the agent should have a full access to the platform
        """
        return CreateAgentTask(json_repr)

    def __init__(self, agent: Dict[str, Any]) -> None:
        """Initialize a new :class:`CreateAgentTask` with the given JSON object."""
        super().__init__()
        self.agent = agent

    async def execute(self, agent: APIAgent) -> AID:
        """
        Ask the AMS to create an initialize a new agent in the platform.

        :param agent: the :class:`APIAgent` that executes the request
        :return: the AID of the created agent
        :raise Exception: the AMS refuse to create the agent or fail at it
        """
        req: ACLMessage = (
            ACLMessage.Builder()
            .performative(Performative.REQUEST)
            .conversation_id(str(uuid4()))
            .receiver(AID(f"ams@{agent.aid.hap_name}"))
            .content(
                [
                    AMSService.CREATE_AGENT_FUNC,
                    AgentCreationDescription(
                        self.agent["class_name"],
                        self.agent["agent_name"],
                        self.agent["args"],
                        self.agent["is_service"],
                    ),
                ]
            )
            .build()
        )
        agent.send(req)

        # Wait response
        agree_or_refuse = await agent.receive(MT_CONVERSATION_ID(req.conversation_id))
        if agree_or_refuse.acl_message.performative == Performative.REFUSE:
            raise Exception(agree_or_refuse.acl_message.content)

        # Wait result
        result = await agent.receive(MT_CONVERSATION_ID(req.conversation_id))
        if result.acl_message.performative == Performative.FAILURE:
            raise Exception(result.acl_message.content)

        return result.acl_message.content[1]  # type: ignore


class GetAgentsTask(Task):
    """
    A task that queries the AMS about agents in the platform.

    Two filters are available:

    - state: if set, filters out agent that have a different state from the one provided
    - name: filters out agents that don't have the provided string in their short name

    On a successful execution, the task returns a list of :class:`piaf.service.AMSAgentDescription`.
    """

    @staticmethod
    def from_json(json_repr: Dict[str, Any]) -> GetAgentsTask:
        """
        Creates a new :class:`GetAgentsTask` from the given JSON object.

        It uses only the `filters` field, which must contain two entries:

        - `state`: if not `None`, must be the string representation of one of the constants in :class:`piaf.agent.AgentState`
        - `name`: a string to filter out agents by their short name

        """
        return GetAgentsTask(json_repr["filters"])

    def __init__(self, filters: Dict[str, Any]) -> None:
        """
        Initialize a new :class:`GetAgentsTask` instance.

        :param filters: a JSON object that contains the two required fields
        """
        super().__init__()
        self.filters = filters

    async def execute(self, agent: APIAgent) -> List[AMSAgentDescription]:
        """
        Ask the AMS the list of agents in the platform an apply filters.

        :param agent: the APIAgent executing the task
        :return: a list of :class:`piaf.service.AMSAgentDescription`
        :raise Exception: the AMS refuse to perform the request or fail at it
        """
        state = (
            None if self.filters["state"] is None else AgentState[self.filters["state"]]
        )
        req: ACLMessage = (
            ACLMessage.Builder()
            .performative(Performative.REQUEST)
            .conversation_id(str(uuid4()))
            .receiver(AID(f"ams@{agent.aid.hap_name}"))
            .content(
                [
                    AMSService.SEARCH_FUNC,
                    AMSAgentDescription(state=state),  # Filter state
                ]
            )
            .build()
        )
        agent.send(req)

        # Wait response
        agree_or_refuse = await agent.receive(MT_CONVERSATION_ID(req.conversation_id))
        if agree_or_refuse.acl_message.performative == Performative.REFUSE:
            raise Exception(agree_or_refuse.acl_message.content)

        # Wait result
        result = await agent.receive(MT_CONVERSATION_ID(req.conversation_id))
        if result.acl_message.performative == Performative.FAILURE:
            raise Exception(result.acl_message.content)

        # Filter using 'name'
        agents: List[AMSAgentDescription] = result.acl_message.content[1]
        return [
            agent for agent in agents if self.filters["name"] in agent.name.short_name
        ]


class ChangeAgentStateTask(Task):
    """
    A task that can manipulate the state of an agent through the AMS.

    It asks the AMS to update the agent's state and return nothing on success.
    """

    @staticmethod
    def from_json(json_repr: Dict[str, Any]) -> ChangeAgentStateTask:
        """
        Initialize a new :class:`ChangeAgentStateTask` object from the given JSON object.

        It expects the following fields:

        - name: the agent's short name
        - state: the string representation of the desired state

        """
        return ChangeAgentStateTask(json_repr["name"], json_repr["state"])

    def __init__(self, name: str, state: str) -> None:
        """
        Initialize the new :class:`ChangeAgentStateTask` instance.

        :param name: the agent's short name
        :param state: the new state, as a string
        """
        super().__init__()
        self._name = name
        self._state = state

    async def execute(self, agent: APIAgent) -> None:
        """
        Call the MODIFY function of the AMS in order to change the agent's state.

        :param agent: the agent executing the task.
        :raise Exception: the AMS refused or failed to modify the state
        """
        req: ACLMessage = (
            ACLMessage.Builder()
            .performative(Performative.REQUEST)
            .conversation_id(str(uuid4()))
            .receiver(AID(f"ams@{agent.aid.hap_name}"))
            .content(
                [
                    AMSService.MODIFY_FUNC,
                    AMSAgentDescription(
                        name=AID(f"{self._name}@{agent.aid.hap_name}"),
                        state=AgentState[self._state],
                    ),
                ]
            )
            .build()
        )
        agent.send(req)

        # Wait response
        agree_or_refuse = await agent.receive(MT_CONVERSATION_ID(req.conversation_id))
        if agree_or_refuse.acl_message.performative == Performative.REFUSE:
            raise Exception(agree_or_refuse.acl_message.content)

        # Wait result (expect an inform)
        result = await agent.receive(MT_CONVERSATION_ID(req.conversation_id))
        if result.acl_message.performative == Performative.FAILURE:
            raise Exception(result.acl_message.content)


class StopPlatformTask(Task):
    """
    A task to stop the platform as soon as possible.

    The agent will close its Redis connection and make the platform stop. Both actions are spawned in a dedicated task to be executed later.
    """

    @staticmethod
    def from_json(ignored: Dict[str, Any]) -> StopPlatformTask:
        """
        Initialize a new :class:`StopPlatformTask` object from the given JSON object.

        It expects no dedicated fields.
        """
        return StopPlatformTask()

    async def execute(self, agent: APIAgent) -> None:
        """Schedule the Redis connection closing and the platform death."""
        asyncio.create_task(agent._platform.stop())


class RetrieveAgentMemoryTask(Task):
    """
    A task to export a snapshot of an agent's memory.

    Only public, non-callable field are included.
    """

    @staticmethod
    def from_json(json_repr: Dict[str, Any]) -> RetrieveAgentMemoryTask:
        """
        Initialize a new :class:`RetrieveAgentMemoryTask` with the provided information.

        It requires one field named `aid`, which itself must contain a `name` field containing the target's fully-qualified name.
        """
        return RetrieveAgentMemoryTask(AID(json_repr["aid"]["name"]))

    def __init__(self, target: AID) -> None:
        """
        Initialize a new :class:`RetrieveAgentMemoryTask`.

        :param target: the AID of the targeted agent
        """
        self.target = target

    async def execute(self, agent: APIAgent) -> Dict[str, Any]:
        try:
            ctx = agent._platform.agent_manager._contexts[self.target]
            target: Agent = ctx.agent

            attrs: Iterable[Tuple[str, Any]] = (
                (attr_name, getattr(target, attr_name)) for attr_name in dir(target)
            )
            memory: Dict[str, Any] = {
                attr: value for (attr, value) in attrs if self._filter_attr(attr, value)
            }
            return {"target": self.target, "memory": memory}

        except KeyError:
            raise Exception(f"No agent with AID={self.target}")

    def _filter_attr(self, attr_name: str, attr_value: Any) -> bool:
        """
        Given an attribute name and value, decide if it should be included in the memory snapshot.

        Excluded attributes are:

        - private attributes (starting with '_')
        - callables (exclude methods)
        - piaf non serializable attributes: state_sync, logger and mailbox

        :param attr_name: name of the attribute
        :param attr_value: value of the attribute
        :return: `True` if the attribute can be included in the memory snapshot, `False` otherwise.
        """
        return (
            (attr_name not in ("state_sync", "logger", "mailbox"))
            and (not attr_name.startswith("_"))
            and (not callable(attr_value))
        )


class SendMessageTask(Task):
    """
    A task that sends, using the given agent's identity, the given message.

    If the sender identity doesn't exist, then the task raises an exception.
    """

    @staticmethod
    def from_json(json_repr: Dict[str, Any]) -> SendMessageTask:
        """
        Initialize a new :class:`SendMessageTask` with the provided information.

        It requires two fields:

        - `sender`, which must contain at least a `name` field containing the full name of the sender agent
        - `msg`, which must contain four fields:
            - `performative`: the message's performative
            - `receivers`: a list of AIDs
            - `conversation_id`: an optional conversation ID
            - `content`: the message content

        """
        msg: Dict[str, Any] = json_repr["msg"]
        sender: AID = AID(json_repr["sender"])
        return SendMessageTask(
            sender,
            ACLMessage.Builder()
            .performative(msg["performative"])
            .receiver([AID(r["name"]) for r in msg["receivers"]])
            .conversation_id(
                msg["conversation_id"]
                if msg.get("conversation_id", None) is not None
                else uuid4()
            )
            .content(msg["content"])
            .build(),
        )

    def __init__(self, sender: AID, msg: ACLMessage) -> None:
        """
        Instantiate a new :class:`SendMessageTask` using the provided sender and the provided message.

        :param sender: which agent is going to send the message
        :param msg: the message to send
        """
        self.sender = sender
        self.msg = msg

    async def execute(self, agent: APIAgent) -> Any:
        try:
            ctx = agent._platform.agent_manager._contexts[self.sender]
            target: Agent = ctx.agent

            target.send(self.msg)

        except KeyError:
            raise Exception(f"No agent with AID={self.sender}")



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['piaf',
 'piaf.api',
 'piaf.api.impl',
 'piaf.comm',
 'piaf.comm.mtp',
 'piaf.examples',
 'piaf.examples.benchmarks',
 'piaf.examples.two_platforms']

package_data = \
{'': ['*']}

install_requires = \
['python-dotenv>=0.21,<0.22']

extras_require = \
{'amqp-mtp': ['aiormq>=6,<7', 'yarl>=1,<2'],
 'webapi': ['aioredis>=2,<3', 'async-timeout>=4,<5', 'fastapi>=0.85,<0.86']}

setup_kwargs = {
    'name': 'piaf',
    'version': '0.2.0',
    'description': 'A FIPA-compliant Agent Platform written in python.',
    'long_description': '# Python Intelligent Agent Framework (piaf)\n\n![pipeline status](https://gitlab.com/ornythorinque/piaf/badges/master/pipeline.svg)\n![coverage report](https://gitlab.com/ornythorinque/piaf/badges/master/coverage.svg?job=test)\n![PyPI - Downloads](https://img.shields.io/pypi/dm/piaf)\n![PyPI - License](https://img.shields.io/pypi/l/piaf)\n![PyPI](https://img.shields.io/pypi/v/piaf)\n![PyPI - Python Version](https://img.shields.io/pypi/pyversions/piaf)\n\nThe aim of piaf is to provide a FIPA-compliant agent framework using Python. It uses **asyncio** to power agents.\n\n## Project status\n\n**The first official release is there!** After almost two years of development, I am now considering piaf is stable enough to start playing with it. I can\'t wait to see the amazing things you will do!\n\nSo, what is next? Well, there are still missing features in piaf and the next version will try to add some of them (I see you FIPA SL!).\n\n## Features\n\nAlthough piaf made some progress, it still needs some love to be fully compliant with the [FIPA specification](http://fipa.org/repository/standardspecs.html).\n\nWe provide some examples to help you understand what is possible to create with the current version, take a look at <https://gitlab.com/ornythorinque/piaf/-/tree/master/src/piaf/examples>.\n\n### Supported features\n\n- AMS (partial, only the query function)\n- DF\n- Communications within a platform\n- Communications between two **piaf platforms** (with some limitations)\n\n### Missing features\n\n- FIPA SL support (only plain Python objects are supported)\n- Federated DF\n- Name resolution\n- "Official" envelope representations (XML, bit-efficient) and MTPs (mainly HTTP, we don\'t plan to support IIOP)\n\n## Documentation\n\nThe full documentation (both user and API) is available here: <https://ornythorinque.gitlab.io/piaf>\nIt will teach you how to install and run your agents.\n\n## Author(s)\n\n* ornythorinque (pierredubaillay@outlook.fr)\n',
    'author': 'Pierre DUBAILLAY',
    'author_email': 'pierredubaillay@outlook.fr',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://gitlab.com/ornythorinque/piaf',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4',
}


setup(**setup_kwargs)



# coding: utf-8
"""
The base module for the Web API part of piaf.

It contains the `FastAPI` application and it imports routes from the :mod:`piaf.api.routes` module.
The application server can be launched using `uvicorn` (or any other production server).
"""
from __future__ import annotations

from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware

import piaf
from piaf.api.dependencies import get_redis_ctrl, get_settings
from piaf.api.routes import app_v1

# Main application
description = """
The Piaf platforms API let you create, manage and delete Piaf platforms.

The API exposes several endpoints related to platforms (create, delete and get running platforms) and agents (create, delete, get state, ...).

Websocket
---------

In addition to the following routes, this API exposes a websocket allowing applications to (un)subscribe to/from event topics. The dedicated route is `ws://{server}/platforms/{ptf_name}/ws` and it is a double-sided websocket carrying JSON data. The client can send the following:

    {
        method: "[un]subscribe",
        topic: ".some.topic"
    }

Once an event is emitted on the subscribed topic, the client receives through the websocket a JSON representation of the `EventRecord`:

    {
        "event": {
            "source": "some_source",
            "type": "some_type",
            "data": ...
        },
        "timestamp": 1663923272,
        "topics": [".platform.agents", ".platform.agents.ams"]
    }

**Warning**: Contrary to how events are dispatched inside the piaf platform, events are not dispatched to topic's parents. It means that listening on `.platform` won't catch events emitted on `.platform.agents` for example.
"""

app = FastAPI(
    title="Piaf platforms management API",
    description=description,
    version=piaf.__version__,
    license_info={"name": "MIT", "url": "https://mit-license.org/"},
    openapi_tags=[
        {"name": "Platform", "description": "Operations with platforms."},
        {"name": "Agent", "description": "Manage agents in a particular platform."},
    ],
)

# Configure CORS
settings = get_settings()

app.add_middleware(
    CORSMiddleware,
    allow_origins=settings.cors_origins,
    allow_credentials=settings.cors_credentials,
    allow_methods=settings.cors_methods,
    allow_headers=settings.cors_headers,
)

# API version 1
app.include_router(app_v1, prefix="/v1")


@app.on_event("startup")
async def initialize_connection() -> None:
    """Initialize the Redis connection used by the WebAPI."""
    await get_redis_ctrl().connect()


@app.on_event("shutdown")
async def close_connection() -> None:
    """Close the Redis connection before stopping the server."""
    await get_redis_ctrl().close()
